-- Capture the Falg v1.0.3
-- by Jon Irons (jonirons@gmail.com)
-- Based on a True Story by Darren Watts

ALERT_SOUND = "juggernaut warning"
PING_SOUND = "computer login"
PING_INTERVAL = 30 * 1 -- dropped flags ping every n seconds
AUTO_RETURN = 30 * 25 -- seconds until a dropped flag returns on its own

ALERT_FADERS = {
   got = {
      we = "burn green",
      they = "burn cyan"
   },
   scored = {
      we = "big white",
      they = "big negative"
   },
   dropped = {
      we = "dodge purple",
      they = "dodge yellow"
   },
   returned = {
      we = "tint blue",
      they = "tint gross"
   }
}

Triggers = {}
BasePolygons = {}
ValidTeams = {}
TeamFlags = {}
TeamBeacons = {}
TeamScores = {}
TeamHUD = {}
TeamStates = {}
Flags = {}

Icons = {}

HUDColors = {}
HUDColors.red = "red"
HUDColors.blue = "blue"

Colors = {}
Colors.red = [[
3
#5e2605
&ff0000
%111111
]]
Colors.blue = [[
3
#5E2605
&0000FF
%111111
]]

Flags.red = "uplink chip"
Flags.blue = "key"

CTF = Game.type == "capture the flag" and not Game.time_remaining

function Triggers.init()

   if CTF then
      populateColorTables()

      for polygon in Polygons() do
	 detectBasePolygon(polygon)
      end

      for index, team in ipairs(ValidTeams) do
	 spawnFlag(team)
      end
      
      for item in Items() do
	 item._age = 0
      end

      for player in Players() do
	 teamUp(player)
      end
   end

end

function Triggers.idle()
   if CTF then

      local teamIndex, team
      for index, team in ipairs(ValidTeams) do
	 if TeamFlags[team.index] == team then
	    spawnFlag(team)
	 end
	 updateHUD(team)
      end

      for item in Items() do
	 ageItem(item)
	 flagPing(item)
	 for color, type in pairsByKeys(Flags) do
	    if TeamFlags[PlayerColors[color].index] and item.type == type
	       and item._age > AUTO_RETURN then
	       autoReturn(item, PlayerColors[color])
	    end
	 end
      end

      for player in Players() do
	 
	 killChangedPlayer(player)
	 reviveChangedPlayer(player)
	 detectCarrier(player)
	 showBeacon(player)
	 showHUD(player)

      end

   else
      if Game.ticks == 0 then
	 for player in Players() do
	    player:print("try again using the CTF game type with a score limit")
	 end
      end

   end
end

function Triggers.got_item(type, player)
   local owner = ownerOfFlag(type)
   if owner then
      if owner == player.team then
	 player.items[type] = 0
	 TeamFlags[player.team.index] = player.team
	 local carrying = hasAFlag(player)
	 if carrying
	    and player.polygon == BasePolygons[player.team.index]
	 then
	    scoreAgainst(player, carrying)
	 else
	    returnFlagToBase(player)
	 end
      else
	 giveEnemyFlag(player, owner)
      end
   end
end

function Triggers.item_created(item)
   item._age = 0
   local flag = ownerOfFlag(item.type)
   if flag and TeamFlags[flag.index] then
      playAll(ALERT_SOUND)
      TeamStates[flag.index] = "dropped"
      for p in Players() do
	 p:print(tostring(flag).."'s flag dropped")
	 if p.team == flag then
	    screen(p, "dropped", "we")
	 else
	    screen(p, "dropped", "they")
	 end
      end
      local beacon = {}
      beacon.x, beacon.y = item.x, item.y
      TeamBeacons[flag.index] = beacon
   end
end

function Triggers.player_revived(player)
   player._needsToLive = false
end

function populateColorTables()
   for color in PlayerColors() do
      BasePolygons[color.index] = false
      TeamBeacons[color.index] = nil
      TeamScores[color.index] = 0
      TeamHUD[color.index] = {}
      TeamStates[color.index] = "base"
   end
end


function detectBasePolygon(polygon)
   if polygon.type == "base" then
      for index, value in ipairs(BasePolygons) do
	 if polygon.permutation == index then
	    BasePolygons[index] = polygon
	    table.insert(ValidTeams, PlayerColors[index])
	 end
      end
   end
end

function teamUp(player)
    if not BasePolygons[player.team.index] then
       player.team = randomValidTeam()
       player._needsToDie = true
    end
   player.color = player.team
end

function randomValidTeam(player)
   return ValidTeams[Game.global_random(#ValidTeams - 1) + 1]
end

function spawnFlag(team)
   local polygon = BasePolygons[team.index]
   TeamFlags[team.index] = nil
   TeamBeacons[team.index] = nil
   TeamStates[team.index] = "base"
   local flag = Items.new(polygon.x, polygon.y, 0,
			  polygon, Flags[tostring(team)])
end

function ownFlag(player, type)
   return Flags[tostring(player.team)] == type
end

function teamPrint(team, message)
   for player in Players() do
      if player.team == team then
	 player:play_sound(ALERT_SOUND, 1)
	 player:print(message)
      end
   end
end

function ownerOfFlag(item)
   for color, type in pairsByKeys(Flags) do
      if item == type then
	 return PlayerColors[color]
      end
   end
   return nil
end

function giveEnemyFlag(player, ownerTeam)
   teamPrint(ownerTeam, tostring(player.team).." has your flag!")
   teamPrint(player.team, player.name.." has "
		..tostring(ownerTeam).."'s flag!")
   for p in Players() do
      if player.team == p.team then
	 screen(p, "got", "we")
      else
	 screen(p, "got", "they")
      end
   end
   player.items[Flags[tostring(ownerTeam)]] = 1
   TeamFlags[ownerTeam.index] = player.team
   TeamStates[ownerTeam.index] = "stolen"
end

function hasAFlag(player)
   for color, type in pairsByKeys(Flags) do
      if player.items[type] > 0 then
	 return PlayerColors[color]
      end
   end
   return nil
end

function removeAllFlags(player)
   for color, type in pairsByKeys(Flags) do
      if player.items[type] > 0 then
	 TeamFlags[PlayerColors[color].index] = PlayerColors[color]
	 player.items[type] = 0
      end
   end
end

function scoreAgainst(player, team)
   removeAllFlags(player)
   player.points = player.points + 1
   TeamScores[player.team.index] = TeamScores[player.team.index] + 1
   teamPrint(player.team, "your team scored against "..tostring(team).."!")
   teamPrint(team, tostring(player.team).." scored against you!")
   playAll(ALERT_SOUND)
   for p in Players() do
	 if p.team == player.team then
	    screen(p, "scored", "we")
	 else
	    screen(p, "scored", "they")
	 end
      end
end

function returnFlagToBase(player)
   local team = player.team
   if "dropped" == TeamStates[team.index] then
      playAll(ALERT_SOUND)
      Players.print(player.name.." returned "
		       ..tostring(team).."'s flag to base.")
      for p in Players() do
	 if p.team == team then
	    screen(p, "returned", "we")
	 else
	    screen(p, "returned", "they")
	 end
      end
   end
end

function ageItem(item)
   item._age = item._age + 1
end

function autoReturn(item, team)
   item:delete()
   playAll(ALERT_SOUND)
   spawnFlag(team)
   Players.print(tostring(team).."'s flag returned to base automatically.")
   for p in Players() do
	 if p.team == team then
	    screen(p, "returned", "we")
	 else
	    screen(p, "returned", "they")
	 end
      end
end

function screen(player, action, who)
   player:fade_screen(ALERT_FADERS[action][who])
end

function detectCarrier(player)	 
   local flag = hasAFlag(player)
   if flag then
      player.color = "white"
      local beacon = {}
      beacon.x, beacon.y = player.x, player.y
      TeamBeacons[flag.index] = beacon
   else
      player.color = player.team
   end
end

function showBeacon(player)
   if TeamBeacons[player.team.index] then
      player.compass.beacon = true
      player.compass.lua = true
      player.compass.x, player.compass.y =
	 TeamBeacons[player.team.index].x, TeamBeacons[player.team.index].y
   else
      player.compass.beacon = false
   end
end

function playAll(sound)
   for player in Players() do
      player:play_sound(sound, 1)
   end
end

function flagPing(item)
   for color, type in pairsByKeys(Flags) do
      if item.type == type and (item._age % PING_INTERVAL) < 1
      and TeamFlags[PlayerColors[color].index]
      then
	 item:play_sound(PING_SOUND)
      end
   end
end

function updateHUD(team)
   local HUD = {}
   local score = TeamScores[team.index]
   local max = Game.kill_limit
   HUD.scoreString = score.." / "..max
   HUD.icon = Colors[tostring(team)]..Icons[TeamStates[team.index]]
   TeamHUD[team.index] = HUD
end

function showHUD(player)
   player.overlays[0]:clear()
   player.overlays[0].color = HUDColors[tostring(player.team)]
   player.overlays[0].icon = TeamHUD[player.team.index].icon
   player.overlays[0].text = TeamHUD[player.team.index].scoreString

   local overlay = 1

   for index, team in ipairs(ValidTeams) do
      if team ~= player.team and overlay < 6 then
	 player.overlays[overlay]:clear()
	 player.overlays[overlay].color = HUDColors[tostring(team)]
	 player.overlays[overlay].icon = TeamHUD[team.index].icon
	 player.overlays[overlay].text = TeamHUD[team.index].scoreString
	 overlay = overlay + 1
      end
   end
end

function killChangedPlayer(player)
   if player._needsToDie then
      player:damage(450, "suffocation")
      player:print("that's what you get for picking a dumb team color")
      player._needsToDie = false
      player._needsToLive = true
   end
end

function reviveChangedPlayer(player)
   if player.dead and player._needsToLive then
      player.action_flags.action_trigger = true
   end
end

--[[
   taken from Programming in Lua; needed to ensure
   order of iteration on maps
]]
function pairsByKeys(t)
   local a = {}
   for n in pairs(t) do table.insert(a, n) end
   table.sort(a)
   local i = 0
   local iterator = function()
      i = i + 1
      if a[i] == nil then return nil
      else return a[i], t[a[i]]
      end
   end
   return iterator
end

Icons.stolen = [[
%##&%%%%%%%&&&&%
%##&&&%%%&&&&&&%
%##&&&&&&&&&&&&%
%##&&&&&&&&&&&&%
###&&&&&&&&&&&&%
##&&&&&&&&&&&&&%
##&&&&&&&&&&&&&%
##%%&&&&&&&%%%%%
##%%%%&&&%%%%%%%
##%%%%%%%%%%%%%%
##%%%%%%%%%%%%%%
##%%%%%%%%%%%%%%
#%%%%%%%%%%%%%%%
#%%%%%%%%%%%%%%%
#%%%%%%%%%%%%%%%
#%%%%%%%%%%%%%%%
]]

Icons.dropped = [[
&&%%%%%%%%%%%%&&
&&&%%%%%%%%%%&&&
%&&&%%%%%%%%&&&%
%%&&&%%%%%%&&&%%
%%%&&&%%%%&&&%%%
%%%%&&&%%&&&%%%%
%%%%%&&&&&&%%%%%
%%%%%%&&&&%%%%%%
%%%%%%&&&&%%%%%%
%%%%%&&&&&&%%%%%
%%%%&&&%%&&&%%%%
%%%&&&%%%%&&&%%%
%%&&&%%%%%%&&&%%
%&&&%%%%%%%%&&&%
&&&%%%%%%%%%%&&&
&&%%%%%%%%%%%%&&
]]

Icons.base = [[
%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%
&&&&&&&&&&&&&&&&
&&&&&&&&&&&&&&&&
&&&&&&&&&&&&&&&&
&&&&&&&&&&&&&&&&
&&&&&&&&&&&&&&&&
&&&&&&&&&&&&&&&&
&&&&&&&&&&&&&&&&
&&&&&&&&&&&&&&&&
%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%
]]
